# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")

TEST_DATA_PATH = Path(__file__).parent / "data" / "semper"


# Reference data:
data_signal = np.arange(27, dtype=np.float32).reshape((3, 3, 3)) / 2.0
data_image = np.arange(16, dtype=np.float32).reshape((4, 4)) / 2.0
data_spectrum = np.arange(10, dtype=np.float32) / 2.0
data_image_byte = np.arange(25, dtype=np.byte).reshape(
    (5, 5)
)  # Odd dim. tests strange read/write
data_image_int16 = np.arange(16, dtype=np.int16).reshape((4, 4))
data_image_int32 = np.arange(16, dtype=np.int32).reshape((4, 4))
data_image_complex = (data_image_int32 + 1j * data_image).astype(np.complex64)
test_title = "This is a test!"


def test_writing_unsupported_data_type(tmp_path):
    data = np.arange(5 * 10).reshape((5, 10))
    s = hs.signals.BaseSignal(data.astype("int64"))
    with pytest.raises(IOError) as cm:
        s.save(filename=tmp_path / "test_writing_unsupported_data_type.unf")
        cm.match("The SEMPER file format does not support int64 data type")


def test_writing_loading_metadata(tmp_path):
    data = np.arange(5 * 10).reshape((5, 10)).astype(np.int8)
    s = hs.signals.BaseSignal(data)
    s.metadata.set_item("General.date", "2016-08-06")
    s.metadata.set_item("General.time", "11:55:00")
    fname = tmp_path / "test_write_with_metadata.unf"
    s.save(fname)
    s2 = hs.load(fname)
    np.testing.assert_allclose(s.data, s2.data)
    assert s.metadata.General.date == s2.metadata.General.date
    assert s.metadata.General.time == s2.metadata.General.time


def test_signal_3d_loading():
    signal = hs.load(TEST_DATA_PATH / "example_signal_3d.unf")
    np.testing.assert_equal(signal.data, data_signal)
    np.testing.assert_equal(signal.original_metadata.IFORM, 2)  # float
    assert isinstance(signal, hs.signals.BaseSignal)


def test_image_2d_loading():
    signal = hs.load(TEST_DATA_PATH / "example_image_2d.unf")
    np.testing.assert_equal(signal.data, data_image)
    np.testing.assert_equal(signal.original_metadata.IFORM, 2)  # float
    assert isinstance(signal, hs.signals.Signal2D)


def test_spectrum_1d_loading():
    signal = hs.load(TEST_DATA_PATH / "example_spectrum_1d.unf")
    np.testing.assert_equal(signal.data, data_spectrum)
    np.testing.assert_equal(signal.original_metadata.IFORM, 2)  # float
    assert isinstance(signal, hs.signals.Signal1D)


def test_image_byte_loading():
    signal = hs.load(TEST_DATA_PATH / "example_image_byte.unf")
    np.testing.assert_equal(signal.data, data_image_byte)
    np.testing.assert_equal(signal.original_metadata.IFORM, 0)  # byte
    assert isinstance(signal, hs.signals.Signal2D)


def test_image_int16_loading():
    signal = hs.load(TEST_DATA_PATH / "example_image_int16.unf")
    np.testing.assert_equal(signal.data, data_image_int16)
    np.testing.assert_equal(signal.original_metadata.IFORM, 1)  # int16
    assert isinstance(signal, hs.signals.Signal2D)


def test_image_int32_loading():
    signal = hs.load(TEST_DATA_PATH / "example_image_int32.unf")
    np.testing.assert_equal(signal.data, data_image_int32)
    np.testing.assert_equal(signal.original_metadata.IFORM, 4)  # int32
    assert isinstance(signal, hs.signals.Signal2D)


def test_image_complex_loading():
    signal = hs.load(TEST_DATA_PATH / "example_image_complex.unf")
    np.testing.assert_equal(signal.data, data_image_complex)
    np.testing.assert_equal(signal.original_metadata.IFORM, 3)  # complex
    assert isinstance(signal, hs.signals.ComplexSignal)


def test_with_title_loading():
    signal = hs.load(TEST_DATA_PATH / "example_with_title.unf")
    np.testing.assert_equal(signal.data, data_image)
    np.testing.assert_equal(signal.original_metadata.IFORM, 2)  # float
    np.testing.assert_equal(signal.metadata.General.title, test_title)
    assert isinstance(signal, hs.signals.Signal2D)


def test_no_label_loading():
    signal = hs.load(TEST_DATA_PATH / "example_no_label.unf")
    np.testing.assert_equal(signal.data, data_image)
    np.testing.assert_equal(signal.original_metadata.ILABEL, 0)
    assert isinstance(signal, hs.signals.Signal2D)


def test_save_and_read_image(tmp_path):
    signal_ref = hs.signals.Signal2D(data_image)
    signal_ref.metadata.General.title = test_title
    signal_ref.save(tmp_path / "example_temp.unf", overwrite=True)
    signal = hs.load(tmp_path / "example_temp.unf")
    np.testing.assert_equal(signal.data, signal_ref.data)
    np.testing.assert_equal(signal.metadata.General.title, test_title)
    assert isinstance(signal, hs.signals.Signal2D)


def test_save_and_read_byte(tmp_path):
    signal_ref = hs.signals.Signal2D(data_image_byte)
    signal_ref.metadata.General.title = test_title
    signal_ref.save(
        tmp_path / "example_temp.unf",
    )
    signal = hs.load(tmp_path / "example_temp.unf")
    np.testing.assert_equal(signal.data, signal_ref.data)
    np.testing.assert_equal(signal.metadata.General.title, test_title)
    assert isinstance(signal, hs.signals.Signal2D)



# -*- coding: utf-8 -*-
# Copyright 2025 CEOS GmbH
# Copyright 2022-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/>.
"""Implements an unpickler, which restricts the allowed modules which pickle can load."""

import pickle
from logging import getLogger

import numpy

from rsciio.utils.file import inspect_npy_bytes


class InvalidPickleError(Exception):
    pass


class RestrictedUnpickler(pickle.Unpickler):
    """
    By overriding `find_class` limits the modules which can be loaded onto the
    pickle stack by the following `white_list`:
    - numpy._core.multiarray._reconstruct
    - numpy._core.multiarray.scalar
    - numpy.ndarray
    - numpy.dtype
    - _codecs.encode
    """

    # The two almost identical NumPy API `core` entries cover older
    # versions (<1.26) and were tested up to NumPy 2.3.
    # Might need additional adaptation for later NumPy >2.3 versions.
    white_list = {
        "numpy._core.multiarray": ["_reconstruct", "scalar"],
        "numpy.core.multiarray": ["_reconstruct", "scalar"],  # numpy <1.26
        "numpy": ["ndarray", "dtype"],
        "_codecs": [
            "encode",
        ],
    }

    def find_class(self, module, name):
        if (module in self.white_list) and (name in self.white_list[module]):
            getLogger(__name__).info(
                "find_class: module = {}, name = {} TRUSTED".format(module, name)
            )
            return super().find_class(module, name)
        else:
            getLogger(__name__).info(
                "find_class: module = {}, name = {} FAILED".format(module, name)
            )
            raise InvalidPickleError(
                f"Invalid names in pickle detected:\n`{module}`, `{name}`"
            )


def read_pickled_array(fp):
    """
    Read pickled data from a NPY file using the RestrictedUnpickler.

    Parameters
    ----------
    fp : file_like object
        The NPY file to read the pickled data from.

    Returns
    -------
    array : ndarray
        The data read from the file.
    """
    # advances the byte stream past the numpy header
    _, _, dtype_str = inspect_npy_bytes(fp)
    dtype = numpy.dtype(dtype_str)
    if not dtype.hasobject:
        raise ValueError("File does not contain pickled data.")

    return RestrictedUnpickler(fp).load()



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import datetime
import logging
import os
import warnings

import dateutil
import numpy as np
import skimage

from rsciio._docstrings import (
    CHUNKS_READ_DOC,
    ENDIANESS_DOC,
    FILENAME_DOC,
    LAZY_DOC,
    RETURNS_DOC,
    SHOW_PROGRESSBAR_DOC,
    SIGNAL_DOC,
)
from rsciio.utils import file
from rsciio.utils._array import dict2sarray, sarray2dict
from rsciio.utils._context_manager import get_progress_bar_context_manager
from rsciio.utils._date_time import (
    datetime_to_serial_date,
    serial_date_to_ISO_format,
)
from rsciio.utils._deprecated import endianess_keyword_deprecation
from rsciio.utils._dictionary import DTBox
from rsciio.utils._skimage_exposure import rescale_intensity

_logger = logging.getLogger(__name__)


magics = [0x0102]


mapping = {
    "blockfile_header.Beam_energy": (
        "Acquisition_instrument.TEM.beam_energy",
        lambda x: x * 1e-3,
    ),
    "blockfile_header.Camera_length": (
        "Acquisition_instrument.TEM.camera_length",
        lambda x: x * 1e-4,
    ),
    "blockfile_header.Scan_rotation": (
        "Acquisition_instrument.TEM.rotation",
        lambda x: x * 1e-2,
    ),
}


def get_header_dtype_list(endianness="<"):
    end = endianness
    dtype_list = (
        [
            ("ID", (bytes, 6)),
            ("MAGIC", end + "u2"),
            ("Data_offset_1", end + "u4"),  # Offset VBF
            ("Data_offset_2", end + "u4"),  # Offset DPs
            ("UNKNOWN1", end + "u4"),  # Flags for ASTAR software?
            ("DP_SZ", end + "u2"),  # Pixel dim DPs
            ("DP_rotation", end + "u2"),  # [degrees ( * 100 ?)]
            ("NX", end + "u2"),  # Scan dim 1
            ("NY", end + "u2"),  # Scan dim 2
            ("Scan_rotation", end + "u2"),  # [100 * degrees]
            ("SX", end + "f8"),  # Pixel size [nm]
            ("SY", end + "f8"),  # Pixel size [nm]
            ("Beam_energy", end + "u4"),  # [V]
            ("SDP", end + "u2"),  # Pixel size [100 * ppcm]
            ("Camera_length", end + "u4"),  # [10 * mm]
            ("Acquisition_time", end + "f8"),  # [Serial date]
        ]
        + [("Centering_N%d" % i, "f8") for i in range(8)]
        + [("Distortion_N%02d" % i, "f8") for i in range(14)]
    )

    return dtype_list


def get_default_header(endianness="<"):
    """Returns a header pre-populated with default values."""
    dt = np.dtype(get_header_dtype_list())
    header = np.zeros((1,), dtype=dt)
    header["ID"][0] = "IMGBLO".encode()
    header["MAGIC"][0] = magics[0]
    header["Data_offset_1"][0] = 0x1000  # Always this value observed
    header["UNKNOWN1"][0] = 131141  # Very typical value (always?)
    header["Acquisition_time"][0] = datetime_to_serial_date(
        datetime.datetime.fromtimestamp(86400, dateutil.tz.tzutc())
    )
    # Default to UNIX epoch + 1 day
    # Have to add 1 day, as dateutil's timezones dont work before epoch
    return header


def get_header_from_signal(signal, endianness="<"):
    from rsciio.utils._units import convert_units

    header = get_default_header(endianness)
    if "blockfile_header" in signal["original_metadata"]:
        header = dict2sarray(
            signal["original_metadata"]["blockfile_header"], sarray=header
        )
        note = signal["original_metadata"]["blockfile_header"]["Note"]
    else:
        note = ""
    # The navigation and signal units are 'nm' and 'cm', respectively, so we
    # convert the units accordingly before saving the signal
    axes = signal["axes"]
    sig_axes = [axis for axis in axes if not axis["navigate"]]
    nav_axes = [axis for axis in axes if axis["navigate"]]
    for axis in sig_axes:
        if axis["units"]:
            try:
                axis["scale"] = convert_units(axis["scale"], axis["units"], "cm")
                axis["offset"] = convert_units(axis["offset"], axis["units"], "cm")
            except Exception:
                warnings.warn(
                    "BLO file expects cm units in signal dimensions. "
                    f"Existing units, {axis['units']} could not be converted; saving "
                    "axes scales as is. Beware that scales "
                    "will likely be incorrect in the file.",
                    UserWarning,
                )
        else:
            warnings.warn(
                "BLO file expects cm units in signal dimensions. "
                f"The {axis['name']} does not have units; saving "
                "axes scales as is. Beware that scales "
                "will likely be incorrect in the file.",
                UserWarning,
            )
    for axis in nav_axes:
        if axis["units"]:
            try:
                axis["scale"] = convert_units(axis["scale"], axis["units"], "nm")
                axis["offset"] = convert_units(axis["offset"], axis["units"], "nm")
            except Exception:
                warnings.warn(
                    "BLO file expects nm units in navigation dimensions. "
                    f"Existing units, {axis['units']} could not be converted; saving "
                    "axes scales as is. Beware that scales "
                    "will likely be incorrect in the file.",
                    UserWarning,
                )
        else:
            warnings.warn(
                "BLO file expects nm units in navigation dimensions. "
                f"The {axis['name']} does not have units; saving "
                "axes scales as is. Beware that scales "
                "will likely be incorrect in the file.",
                UserWarning,
            )

    if len(nav_axes) == 2:
        NX = nav_axes[1]["size"]
        NY = nav_axes[0]["size"]
        SX = nav_axes[1]["scale"]
        SY = nav_axes[0]["scale"]
    elif len(nav_axes) == 1:
        NX = nav_axes[0]["size"]
        NY = 1
        SX = nav_axes[0]["scale"]
        SY = SX
    elif len(nav_axes) == 0:
        NX = NY = SX = SY = 1
    else:
        raise ValueError(
            "Number of navigation axes has to be 0, 1 or 2"
        )  # pragma: no cover

    DP_SZ = [axis["size"] for axis in sig_axes][::-1]
    if DP_SZ[0] != DP_SZ[1]:
        raise ValueError("Blockfiles require signal shape to be square!")
    DP_SZ = DP_SZ[0]
    SDP = 100.0 / sig_axes[1]["scale"]

    offset2 = NX * NY + header["Data_offset_1"][0]
    # Based on inspected files, the DPs are stored at 16-bit boundary...
    # Normally, you'd expect word alignment (32-bits) ¯\_(°_o)_/¯
    offset2 += offset2 % 16

    header = dict2sarray(
        {
            "NX": NX,
            "NY": NY,
            "DP_SZ": DP_SZ,
            "SX": SX,
            "SY": SY,
            "SDP": SDP,
            "Data_offset_2": offset2,
        },
        sarray=header,
    )
    return header, note


@endianess_keyword_deprecation
def file_reader(filename, lazy=False, chunks="auto", endianness="<"):
    """
    Read a blockfile.

    Parameters
    ----------
    %s
    %s
    %s
    %s

    %s
    """

    _logger.debug("Reading blockfile: %s" % filename)
    metadata = {}
    f = open(filename, "rb")
    _logger.debug("File opened")

    # Get header
    header = np.fromfile(f, dtype=get_header_dtype_list(endianness), count=1)
    if header["MAGIC"][0] not in magics:
        warnings.warn(
            "Blockfile has unrecognized header signature. "
            "Will attempt to read, but correctness not guaranteed!",
            UserWarning,
        )
    header = sarray2dict(header)
    # Convert the header value to python int to avoid potential overflow issues on 32-bit systems
    note = f.read(int(header["Data_offset_1"]) - f.tell())
    # It seems it uses "\x00" for padding, so we remove it
    try:
        header["Note"] = note.decode("latin1").strip("\x00")
    except Exception:
        # Not sure about the encoding so, if it fails, we carry on
        _logger.warning(
            "Reading the Note metadata of this file failed. "
            "You can help improving "
            "HyperSpy by reporting the issue in "
            "https://github.com/hyperspy/hyperspy"
        )
    _logger.debug("File header: " + str(header))
    navigation_shape = (int(header["NY"]), int(header["NX"]))
    signal_shape = (int(header["DP_SZ"]),) * 2
    if header["SDP"]:
        SDP = 100.0 / header["SDP"]
    else:
        SDP = 1  #  Set default scale to 1
    original_metadata = {"blockfile_header": header}

    # Get data:

    # TODO A Virtual BF/DF is stored first, may be loaded as navigator in future
    # offset1 = header['Data_offset_1']
    # f.seek(offset1)
    # navigator = np.fromfile(f, dtype=endianness+"u1", shape=(NX, NY)).T

    # Then comes actual blockfile
    # header["Data_offset_2"] is np.uint32. Converting to int to avoid potential overflow issues on 32-bit systems
    offset2 = int(header["Data_offset_2"])
    # Every frame is preceded by a 6 byte sequence
    # (AA 55, and then a 4 byte integer specifying frame number)
    if not lazy:
        frame_dtype = np.dtype(
            [
                ("header", np.dtype(endianness + "u1"), 6),
                ("data", np.dtype(endianness + "u1"), np.prod(signal_shape)),
            ]
        )
        f.seek(offset2)
        data = np.fromfile(f, dtype=frame_dtype)["data"]
        try:
            data = data.reshape(navigation_shape + signal_shape).squeeze()
        except ValueError:
            warnings.warn(
                "Blockfile header dimensions larger than file size! "
                "Will attempt to load by zero padding incomplete frames."
            )
            # Data is stored DP by DP:
            pw = [(0, np.prod(navigation_shape) - data.shape[0]), (0, 0)]
            data = np.pad(data, pw, mode="constant")
            data = data.reshape(navigation_shape + signal_shape).squeeze()
    else:
        frame_dtype = np.dtype(
            [
                ("header", np.dtype(endianness + "u1"), 6),
                ("data", np.dtype(endianness + "u1"), signal_shape),
            ]
        )
        data = file.memmap_distributed(
            filename,
            chunks=chunks,
            offset=offset2,
            shape=navigation_shape,
            dtype=frame_dtype,
            key="data",
        )

    units = ["nm", "nm", "cm", "cm"]
    names = ["y", "x", "dy", "dx"]
    navigate = [True, True, False, False]
    scales = [header["SY"], header["SX"], SDP, SDP]
    date, time, time_zone = serial_date_to_ISO_format(header["Acquisition_time"])
    metadata = {
        "General": {
            "original_filename": os.path.split(filename)[1],
            "date": date,
            "time": time,
            "time_zone": time_zone,
            "notes": header["Note"],
        },
        "Signal": {"signal_type": "diffraction"},
    }
    # Create the axis objects for each axis
    dim = data.ndim
    axes = [
        {
            "size": data.shape[i],
            "index_in_array": i,
            "name": names[i],
            "scale": scales[i],
            "offset": 0.0,
            "units": units[i],
            "navigate": navigate[i],
        }
        for i in range(dim)
    ]

    dictionary = {
        "data": data,
        "axes": axes,
        "metadata": metadata,
        "original_metadata": original_metadata,
        "mapping": mapping,
    }

    f.close()
    return [
        dictionary,
    ]


file_reader.__doc__ %= (
    FILENAME_DOC,
    LAZY_DOC,
    CHUNKS_READ_DOC,
    ENDIANESS_DOC,
    RETURNS_DOC,
)


def file_writer(
    filename,
    signal,
    intensity_scaling=None,
    navigator="navigator",
    show_progressbar=True,
    endianness="<",
):
    """
    Write signal to blockfile.

    Parameters
    ----------
    %s
    %s
    intensity_scaling : str, 2-tuple of float, 2-tuple of int
        If the signal datatype is not :py:class:`numpy.ubyte`, casting to this
        datatype without intensity rescaling results in overflow errors (default behavior)
        This argument provides intensity scaling strategies and the options are:

        - ``'dtype'``: the limits of the datatype of the dataset, e.g. 0-65535 for
          :py:class:`numpy.ushort`, are mapped onto 0-255, respectively. Does not work
          for ``float`` data types.
        - ``'minmax'``: the minimum and maximum in the dataset are mapped to 0-255.
        - ``'crop'``: everything below 0 and above 255 is set to 0 and 255, respectively
        - 2-tuple of `floats` or `ints`: the intensities between these values are
          scaled between 0-255, everything below is 0 and everything above is 255.
    navigator : str or array-like
        A ``.blo`` file also saves a virtual bright field image for navigation.
        This option determines what kind of data is stored for this image.
        By default this is set to ``'navigator'``, which results in using the
        :py:attr:`hyperspy.api.signals.BaseSignal.navigator` attribute if used with HyperSpy.
        Otherwise, it is calculated during saving which can take  some time for large
        datasets. Alternatively, an array-like of the right shape may also be provided.
        If set to None, a zero array is stored in the file.
    %s
    %s
    """
    smetadata = DTBox(signal["metadata"], box_dots=True)
    if intensity_scaling is None:
        # to distinguish from the tuple case
        if signal["data"].dtype != "u1":
            warnings.warn(
                "Data does not have uint8 dtype: values outside the "
                "range 0-255 may result in overflow. To avoid this "
                "use the 'intensity_scaling' keyword argument.",
                UserWarning,
            )
    elif intensity_scaling == "dtype":
        original_scale = skimage.dtype_limits(signal["data"])
        if original_scale[1] == 1.0:
            raise ValueError("Signals with float dtype can not use 'dtype'")
    elif intensity_scaling == "minmax":
        minimum = signal["data"].min()
        maximum = signal["data"].max()
        if signal["attributes"]["_lazy"]:
            import dask

            minimum, maximum = dask.compute(minimum, maximum)
        original_scale = (minimum, maximum)
    elif intensity_scaling == "crop":
        original_scale = (0, 255)
    else:
        # we leave the error checking for incorrect tuples to skimage
        original_scale = intensity_scaling

    header, note = get_header_from_signal(signal, endianness=endianness)
    with open(filename, "wb") as f:
        # Write header
        header.tofile(f)
        # Write header note field:
        if len(note) > int(header["Data_offset_1"][0]) - f.tell():
            note = note[: int(header["Data_offset_1"][0]) - f.tell() - len(note)]
        f.write(note.encode())
        # Zero pad until next data block
        zero_pad = int(header["Data_offset_1"][0]) - f.tell()
        np.zeros((zero_pad,), np.byte).tofile(f)
        # Write virtual bright field
        if navigator is None:
            navigator = np.zeros((signal["data"].shape[0], signal["data"].shape[1]))
        elif isinstance(navigator, str) and (navigator == "navigator"):
            if smetadata.get("_HyperSpy._sig_navigator", False):
                navigator = smetadata["_HyperSpy._sig_navigator.data"]
            else:
                navigator = signal["data"].mean(axis=(-2, -1))
        elif hasattr(navigator, "shape"):
            # Is numpy array-like
            # check that the shape is ok
            if navigator.shape != signal["data"].shape[:2]:
                raise ValueError(
                    "Size of the provided `navigator` does not match the "
                    "navigation dimensions of the dataset."
                )
        else:
            raise ValueError("The `navigator` argument is expected to be array-like")
        if intensity_scaling is not None:
            navigator = rescale_intensity(
                navigator, in_range=original_scale, out_range=np.uint8
            )
        navigator = navigator.astype(endianness + "u1")
        np.asanyarray(navigator).tofile(f)
        # Zero pad until next data block
        if f.tell() > int(header["Data_offset_2"][0]):
            raise ValueError("Signal navigation size does not match data dimensions.")
        zero_pad = int(header["Data_offset_2"][0]) - f.tell()
        np.zeros((zero_pad,), np.byte).tofile(f)
        file_location = f.tell()

    if intensity_scaling is not None:
        array_data = rescale_intensity(
            signal["data"],
            in_range=original_scale,
            out_range=np.uint8,
        )
    else:
        array_data = signal["data"]
    array_data = array_data.astype(endianness + "u1")
    # Write full data stack:
    # We need to pad each image with magic 'AA55', then a u32 serial
    pixels = array_data.shape[-2:]
    records = array_data.shape[:-2]
    record_dtype = [
        ("MAGIC", endianness + "u2"),
        ("ID", endianness + "u4"),
        ("IMG", endianness + "u1", pixels),
    ]
    magics = np.full(records, 0x55AA, dtype=endianness + "u2")
    ids = np.arange(np.prod(records), dtype=endianness + "u4").reshape(records)
    file_memmap = np.memmap(
        filename, dtype=record_dtype, mode="r+", offset=file_location, shape=records
    )
    file_memmap["MAGIC"] = magics
    file_memmap["ID"] = ids
    if signal["attributes"]["_lazy"]:
        with get_progress_bar_context_manager(show_progressbar)():
            array_data.store(file_memmap["IMG"])
    else:
        file_memmap["IMG"] = array_data
    file_memmap.flush()


file_writer.__doc__ %= (
    FILENAME_DOC.replace("read", "write to"),
    SIGNAL_DOC,
    SHOW_PROGRESSBAR_DOC,
    ENDIANESS_DOC,
)



# -*- coding: utf-8 -*-
# Copyright 2007-2025 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import os

import yaml

__all__ = ["IO_PLUGINS"]


def __dir__():
    return sorted(__all__)


class IOPLUGINS(list):
    # Use a class to allow for docstring
    """
    List of available IO plugins.

    Each entry is a dictionary with the following keys:

    - ``'name'``: The name of the plugin.
    - ``'name_aliases'``: A list of alternative names for the plugin.
    - ``'description'``: A brief description of the plugin.
    - ``'full_support'``: A boolean indicating if the plugin has full support.
    - ``'default_extension'``: The default file extension for the plugin.
    - ``'writes'``: A boolean indicating if the plugin supports writing files.
    - ``'non_uniform_axis'``: A boolean indicating if the plugin supports non-uniform axes.
    - ``'api'``: The API module path as a string (e.g., 'rsciio.nexus').

    :meta hide-value:
    """


IO_PLUGINS = IOPLUGINS()

# libyaml C bindings may be missing
loader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)

for sub, _, _ in os.walk(os.path.abspath(os.path.dirname(__file__))):
    _specs_file = os.path.join(sub, "specifications.yaml")
    if os.path.isfile(_specs_file):
        with open(_specs_file, "r") as stream:
            _specs = yaml.load(stream, Loader=loader)
            # for testing purposes
            _specs["api"] = "rsciio.%s" % os.path.split(sub)[1]
            IO_PLUGINS.append(_specs)



from pathlib import Path

from rsciio.utils.path import (
    append2pathname,
    ensure_directory,
    incremental_filename,
    overwrite,
)


def test_append2pathname():
    filename = "data/file.txt"
    to_append = "_std"
    new_filename = append2pathname(filename, to_append)
    assert new_filename == Path("data/file_std.txt")
    assert new_filename.suffix == ".txt"
    assert new_filename.parent == Path("data")


def test_incremental_filename(tmp_path):
    base_filename = tmp_path / "output.txt"
    # Create the first file
    base_filename.touch()
    # Get an incremental filename
    new_filename = incremental_filename(base_filename)
    assert new_filename == tmp_path / "output-1.txt"
    # Create the second file
    new_filename.touch()
    # Get another incremental filename
    next_filename = incremental_filename(base_filename)
    assert next_filename == tmp_path / "output-2.txt"


def test_ensure_no_increment_if_file_not_exists(tmp_path):
    base_filename = tmp_path / "new_output.txt"
    # Ensure the file does not exist
    if base_filename.is_file():
        base_filename.unlink()
    # Get an incremental filename
    new_filename = incremental_filename(base_filename)
    assert new_filename == base_filename
    assert not new_filename.is_file()


def test_ensure_directory(tmp_path):
    dir_path = tmp_path / "new_directory"
    # Ensure the directory does not exist
    if dir_path.is_dir():
        dir_path.rmdir()
    # Use the function to ensure the directory exists
    ensure_directory(dir_path)
    assert dir_path.is_dir()
    # Call again to ensure it does not raise an error if the directory already exists
    ensure_directory(dir_path)


def test_overwrite_file_not_exists(tmp_path):
    # Test with non-existent file
    test_file = tmp_path / "non_existent.txt"

    result = overwrite(test_file)
    assert result is True



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib
import logging
import sys
import time
from pathlib import Path

import dask.array as da
import numpy as np
import pytest

from rsciio.utils._tests import assert_deep_almost_equal
from rsciio.utils._tools import get_file_handle

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
h5py = pytest.importorskip("h5py", reason="h5py not installed")

from hyperspy.axes import (  # noqa: E402
    AxesManager,
    DataAxis,
    FunctionalDataAxis,
    UniformDataAxis,
)
from hyperspy.decorators import lazifyTestClass  # noqa: E402
from hyperspy.exceptions import VisibleDeprecationWarning  # noqa: E402
from hyperspy.misc.test_utils import sanitize_dict as san_dict  # noqa: E402

from rsciio._hierarchical import get_signal_chunks  # noqa: E402

TEST_DATA_PATH = Path(__file__).parent / "data" / "hspy"
TEST_NPZ_DATA_PATH = Path(__file__).parent / "data" / "npz"


if importlib.util.find_spec("zarr") is None:
    zspy_marker = pytest.mark.parametrize("file", ["test.hspy"])
else:
    zspy_marker = pytest.mark.parametrize("file", ["test.hspy", "test.zspy"])


data = np.array(
    [
        4066.0,
        3996.0,
        3932.0,
        3923.0,
        5602.0,
        5288.0,
        7234.0,
        7809.0,
        4710.0,
        5015.0,
        4366.0,
        4524.0,
        4832.0,
        5474.0,
        5718.0,
        5034.0,
        4651.0,
        4613.0,
        4637.0,
        4429.0,
        4217.0,
    ]
)
example1_original_metadata = {
    "BEAMDIAM -nm": 100.0,
    "BEAMKV   -kV": 120.0,
    "CHOFFSET": -168.0,
    "COLLANGLE-mR": 3.4,
    "CONVANGLE-mR": 1.5,
    "DATATYPE": "XY",
    "DATE": "01-OCT-1991",
    "DWELLTIME-ms": 100.0,
    "ELSDET": "SERIAL",
    "EMISSION -uA": 5.5,
    "FORMAT": "EMSA/MAS Spectral Data File",
    "MAGCAM": 100.0,
    "NCOLUMNS": 1.0,
    "NPOINTS": 20.0,
    "OFFSET": 520.13,
    "OPERMODE": "IMAG",
    "OWNER": "EMSA/MAS TASK FORCE",
    "PROBECUR -nA": 12.345,
    "SIGNALTYPE": "ELS",
    "THICKNESS-nm": 50.0,
    "TIME": "12:00",
    "TITLE": "NIO EELS OK SHELL",
    "VERSION": "1.0",
    "XLABEL": "Energy",
    "XPERCHAN": 3.1,
    "XUNITS": "eV",
    "YLABEL": "Counts",
    "YUNITS": "Intensity",
}


class Example1:
    "Used as a base class for the TestExample classes below"

    def test_data(self):
        assert [
            4066.0,
            3996.0,
            3932.0,
            3923.0,
            5602.0,
            5288.0,
            7234.0,
            7809.0,
            4710.0,
            5015.0,
            4366.0,
            4524.0,
            4832.0,
            5474.0,
            5718.0,
            5034.0,
            4651.0,
            4613.0,
            4637.0,
            4429.0,
            4217.0,
        ] == self.s.data.tolist()

    def test_original_metadata(self):
        assert example1_original_metadata == self.s.original_metadata.as_dictionary()


class TestExample1_12(Example1):
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example1_v1.2.hdf5", reader="HSPY")

    def test_date(self):
        assert self.s.metadata.General.date == "1991-10-01"

    def test_time(self):
        assert self.s.metadata.General.time == "12:00:00"


class TestExample1_10(Example1):
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example1_v1.0.hdf5", reader="HSPY")


class TestExample1_11(Example1):
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example1_v1.1.hdf5", reader="HSPY")


class TestLoadingNewSavedMetadata:
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "with_lists_etc.hdf5", reader="HSPY")

    def test_signal_inside(self):
        np.testing.assert_array_almost_equal(
            self.s.data, self.s.metadata.Signal.Noise_properties.variance.data
        )

    def test_empty_things(self):
        assert self.s.metadata.test.empty_list == []
        assert self.s.metadata.test.empty_tuple == ()

    def test_simple_things(self):
        assert self.s.metadata.test.list == [42]
        assert self.s.metadata.test.tuple == (1, 2)

    def test_inside_things(self):
        assert self.s.metadata.test.list_inside_list == [42, 137, [0, 1]]
        assert self.s.metadata.test.list_inside_tuple == (137, [42, 0])
        assert self.s.metadata.test.tuple_inside_tuple == (137, (123, 44))
        assert self.s.metadata.test.tuple_inside_list == [137, (123, 44)]

    @pytest.mark.xfail(reason="dill is not guaranteed to load across Python versions")
    def test_binary_string(self):
        import dill

        # apparently pickle is not "full" and marshal is not
        # backwards-compatible
        f = dill.loads(self.s.metadata.test.binary_string)
        assert f(3.5) == 4.5


class TestSavingMetadataContainers:
    def setup_method(self, method):
        self.s = hs.signals.BaseSignal([0.1])

    @zspy_marker
    def test_save_unicode(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", ["a", "b", "\u6f22\u5b57"])
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert isinstance(s2.metadata.test[0], str)
        assert isinstance(s2.metadata.test[1], str)
        assert isinstance(s2.metadata.test[2], str)
        assert s2.metadata.test[2] == "\u6f22\u5b57"

    @pytest.mark.xfail(reason="osx is slow occasionally")
    @zspy_marker
    def test_save_long_list(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("long_list", list(range(10000)))
        start = time.time()
        fname = tmp_path / file
        s.save(fname)
        end = time.time()
        # It should finish in less that 2 s on CI
        assert end - start < 2.0

    @zspy_marker
    def test_numpy_only_inner_lists(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", [[1.0, 2], ("3", 4)])
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert isinstance(s2.metadata.test, list)
        assert isinstance(s2.metadata.test[0], list)
        assert isinstance(s2.metadata.test[1], tuple)

    @pytest.mark.xfail(sys.platform == "win32", reason="randomly fails in win32")
    @zspy_marker
    def test_numpy_general_type(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", np.array([[1.0, 2], ["3", 4]]))
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        np.testing.assert_array_equal(s2.metadata.test, s.metadata.test)

    @pytest.mark.xfail(sys.platform == "win32", reason="randomly fails in win32")
    @zspy_marker
    def test_list_general_type(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", [[1.0, 2], ["3", 4]])
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert isinstance(s2.metadata.test[0][0], float)
        assert isinstance(s2.metadata.test[0][1], float)
        assert isinstance(s2.metadata.test[1][0], str)
        assert isinstance(s2.metadata.test[1][1], str)

    @pytest.mark.xfail(sys.platform == "win32", reason="randomly fails in win32")
    @zspy_marker
    def test_general_type_not_working(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", (hs.signals.BaseSignal([1]), 0.1, "test_string"))
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert isinstance(s2.metadata.test, tuple)
        assert isinstance(s2.metadata.test[0], hs.signals.Signal1D)
        assert isinstance(s2.metadata.test[1], float)
        assert isinstance(s2.metadata.test[2], str)

    @zspy_marker
    def test_unsupported_type(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", hs.roi.Point2DROI(1, 2))
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert "test" not in s2.metadata

    @zspy_marker
    def test_date_time(self, tmp_path, file):
        s = self.s
        date, time = "2016-08-05", "15:00:00.450"
        s.metadata.General.date = date
        s.metadata.General.time = time
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.metadata.General.date == date
        assert s2.metadata.General.time == time

    @zspy_marker
    def test_general_metadata(self, tmp_path, file):
        s = self.s
        notes = "Dummy notes"
        authors = "Author 1, Author 2"
        doi = "doi"
        s.metadata.General.notes = notes
        s.metadata.General.authors = authors
        s.metadata.General.doi = doi
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.metadata.General.notes == notes
        assert s2.metadata.General.authors == authors
        assert s2.metadata.General.doi == doi

    @zspy_marker
    def test_quantity(self, tmp_path, file):
        s = self.s
        quantity = "Intensity (electron)"
        s.metadata.Signal.quantity = quantity
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.metadata.Signal.quantity == quantity

    @zspy_marker
    def test_save_axes_manager(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", s.axes_manager)
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        # strange because you need the encoding...
        assert isinstance(s2.metadata.test, AxesManager)

    @zspy_marker
    def test_title(self, tmp_path, file):
        s = self.s
        fname = tmp_path / file
        s.metadata.General.title = "__unnamed__"
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.metadata.General.title == ""

    @zspy_marker
    def test_save_empty_tuple(self, tmp_path, file):
        s = self.s
        s.metadata.set_item("test", ())
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        # strange because you need the encoding...
        assert s2.metadata.test == s.metadata.test

    @zspy_marker
    def test_save_bytes(self, tmp_path, file):
        s = self.s
        byte_message = bytes("testing", "utf-8")
        s.metadata.set_item("test", byte_message)
        fname = tmp_path / file
        s.save(fname)
        s2 = hs.load(fname)
        assert s2.metadata.test == s.metadata.test.decode()

    def test_metadata_binned_deprecate(self):
        with pytest.warns(UserWarning, match="Loading old file"):
            s = hs.load(TEST_DATA_PATH / "example2_v2.2.hspy")
        assert s.metadata.has_item("Signal.binned") is False
        assert s.axes_manager[-1].is_binned is False

    def test_metadata_update_to_v3_1(self):
        md = {
            "Acquisition_instrument": {
                "SEM": {"Stage": {"tilt_alpha": 5.0}},
                "TEM": {
                    "Detector": {"Camera": {"exposure": 0.20000000000000001}},
                    "Stage": {"tilt_alpha": 10.0},
                    "acquisition_mode": "TEM",
                    "beam_current": 0.0,
                    "beam_energy": 200.0,
                    "camera_length": 320.00000000000006,
                    "microscope": "FEI Tecnai",
                },
            },
            "General": {
                "date": "2014-07-09",
                "original_filename": "test_diffraction_pattern.dm3",
                "time": "18:56:37",
                "title": "test_diffraction_pattern",
                "FileIO": {
                    "0": {
                        "operation": "load",
                        "hyperspy_version": hs.__version__,
                        "io_plugin": "rsciio.hspy",
                    }
                },
            },
            "Signal": {
                "Noise_properties": {
                    "Variance_linear_model": {"gain_factor": 1.0, "gain_offset": 0.0}
                },
                "quantity": "Intensity",
                "signal_type": "",
            },
            "_HyperSpy": {
                "Folding": {
                    "original_axes_manager": None,
                    "original_shape": None,
                    "signal_unfolded": False,
                    "unfolded": False,
                }
            },
        }
        s = hs.load(TEST_DATA_PATH / "example2_v3.1.hspy")
        # delete timestamp from metadata since it's runtime dependent
        del s.metadata.General.FileIO.Number_0.timestamp
        assert_deep_almost_equal(s.metadata.as_dictionary(), md)


def test_none_metadata():
    s = hs.load(TEST_DATA_PATH / "none_metadata.hdf5", reader="HSPY")
    assert s.metadata.should_be_None is None


def test_rgba16():
    print(TEST_DATA_PATH)  # noqa: T201
    with pytest.warns(VisibleDeprecationWarning):
        #  The binned attribute has been moved from metadata.Signal
        s = hs.load(TEST_DATA_PATH / "test_rgba16.hdf5", reader="HSPY")
    data = np.load(TEST_NPZ_DATA_PATH / "test_rgba16.npz")["a"]
    assert (s.data == data).all()


def test_non_valid_hspy(tmp_path, caplog):
    filename = tmp_path / "testfile.hspy"
    data = np.arange(10)

    with h5py.File(filename, mode="w") as f:
        f.create_dataset("dataset", data=data)

    with pytest.raises(IOError):
        with caplog.at_level(logging.ERROR):
            _ = hs.load(filename)


@zspy_marker
@pytest.mark.parametrize("lazy", [True, False])
def test_nonuniformaxis(tmp_path, file, lazy):
    fname = tmp_path / file
    data = np.arange(10)
    axis = DataAxis(axis=1 / np.arange(1, data.size + 1), navigate=False)
    s = hs.signals.Signal1D(data, axes=(axis.get_axis_dictionary(),))
    if lazy:
        s = s.as_lazy()
    s.save(fname, overwrite=True)
    s2 = hs.load(fname)
    np.testing.assert_array_almost_equal(
        s.axes_manager[0].axis, s2.axes_manager[0].axis
    )
    assert s2.axes_manager[0].is_uniform is False
    assert s2.axes_manager[0].navigate is False
    assert s2.axes_manager[0].size == data.size


@zspy_marker
@pytest.mark.parametrize("lazy", [True, False])
def test_nonuniformFDA(tmp_path, file, lazy):
    fname = tmp_path / file
    data = np.arange(10)
    x0 = UniformDataAxis(size=data.size, offset=1)
    axis = FunctionalDataAxis(expression="1/x", x=x0, navigate=False)
    s = hs.signals.Signal1D(data, axes=(axis.get_axis_dictionary(),))
    if lazy:
        s = s.as_lazy()
    print(axis.get_axis_dictionary())  # noqa: T201
    s.save(fname, overwrite=True)
    s2 = hs.load(fname)
    np.testing.assert_array_almost_equal(
        s.axes_manager[0].axis, s2.axes_manager[0].axis
    )
    assert s2.axes_manager[0].is_uniform is False
    assert s2.axes_manager[0].navigate is False
    assert s2.axes_manager[0].size == data.size


def test_lazy_loading(tmp_path):
    s = hs.signals.BaseSignal(np.empty((5, 5, 5)))
    fname = tmp_path / "tmp.hdf5"
    s.save(fname, overwrite=True, file_format="HSPY")
    shape = (10000, 10000, 100)
    del s
    f = h5py.File(fname, mode="r+")
    s = f["Experiments/__unnamed__"]
    del s["data"]
    s.create_dataset("data", shape=shape, dtype="float64", chunks=True)
    f.close()

    s = hs.load(fname, lazy=True, reader="HSPY")
    assert shape == s.data.shape
    assert isinstance(s.data, da.Array)
    assert s._lazy
    s.close_file()


def test_passing_compression_opts_saving(tmp_path):
    filename = tmp_path / "testfile.hdf5"
    hs.signals.BaseSignal([1, 2, 3]).save(
        filename, compression_opts=8, file_format="HSPY"
    )

    f = h5py.File(filename, mode="r+")
    d = f["Experiments/__unnamed__/data"]
    assert d.compression_opts == 8
    assert d.compression == "gzip"
    f.close()


@zspy_marker
def test_axes_configuration(tmp_path, file):
    fname = tmp_path / file
    s = hs.signals.BaseSignal(np.zeros((2, 2, 2, 2, 2)))
    s.axes_manager.signal_axes[0].navigate = True
    s.axes_manager.signal_axes[0].navigate = True

    s.save(fname, overwrite=True)
    s = hs.load(fname)
    assert s.axes_manager.navigation_axes[0].index_in_array == 4
    assert s.axes_manager.navigation_axes[1].index_in_array == 3
    assert s.axes_manager.signal_dimension == 3


@zspy_marker
def test_axes_configuration_binning(tmp_path, file):
    fname = tmp_path / file
    s = hs.signals.BaseSignal(np.zeros((2, 2, 2)))
    s.axes_manager.signal_axes[-1].is_binned = True
    s.save(fname)

    s = hs.load(fname)
    assert s.axes_manager.signal_axes[-1].is_binned


@lazifyTestClass
class TestPermanentMarkersIO:
    def setup_method(self, method):
        s = hs.signals.Signal2D(np.arange(100).reshape(10, 10))
        self.s = s

    @zspy_marker
    def test_save_permanent_markers(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        m = hs.plot.markers.Points(offsets=np.array([5, 5]))
        s.add_marker(m, permanent=True)
        s.save(filename)

    @zspy_marker
    def test_save_load_empty_metadata_markers(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        m = hs.plot.markers.Points(offsets=np.array([5, 5]))
        m.name = "test"
        s.add_marker(m, permanent=True)
        del s.metadata.Markers.test
        s.save(filename)
        s1 = hs.load(filename)
        assert len(s1.metadata.Markers) == 0

    @zspy_marker
    def test_save_load_permanent_markers(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        x, y = 5, 2
        color = "red"
        sizes = 10
        name = "testname"
        m = hs.plot.markers.Points(offsets=np.array([x, y]), color=color, sizes=sizes)
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        assert len(s1.metadata.Markers) == len(s.metadata.Markers)
        assert s1.metadata.Markers.has_item(name)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_allclose(m1.get_current_kwargs()["offsets"], np.array([x, y]))
        assert m1.get_current_kwargs()["sizes"] == (sizes,)
        # assert m1.get_current_kwargs()["color"] == (color, )
        assert m1.name == name

    @zspy_marker
    def test_save_load_permanent_markers_all_types(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        x1, y1, x2, y2 = 5, 2, 1, 8
        offsets = np.array([x1, y1])
        rng = np.random.default_rng(0)
        segments = rng.random((10, 2, 2)) * 100
        X, Y = np.meshgrid(np.arange(0, 2 * np.pi, 0.2), np.arange(0, 2 * np.pi, 0.2))
        # For arrows
        offsets_arrow = np.column_stack((X.ravel(), Y.ravel()))
        U = np.cos(X).ravel() / 7.5
        V = np.sin(Y).ravel() / 7.5
        C = np.hypot(U, V)
        poylgon = [[1, 1], [20, 20], [1, 20], [25, 5]]
        m0_list = [
            hs.plot.markers.Arrows(offsets_arrow, U, V, C=C),
            hs.plot.markers.Points(offsets),
            hs.plot.markers.HorizontalLines(offsets=(y1,)),
            hs.plot.markers.Lines(segments=segments),
            hs.plot.markers.Rectangles(offsets, widths=(x2 - x1,), heights=(y2 - y1,)),
            hs.plot.markers.Texts(offsets, texts="test"),
            hs.plot.markers.VerticalLines(offsets=(y1,)),
            hs.plot.markers.Polygons(
                verts=[
                    poylgon,
                ]
            ),
        ]
        for m in m0_list:
            s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        assert len(s1.metadata.Markers) == len(s.metadata.Markers)

        markers_dict = s1.metadata.Markers
        m0_dict_list = []
        m1_dict_list = []
        for m in m0_list:
            m0_dict_list.append(san_dict(m._to_dictionary()))
            m1_dict_list.append(
                san_dict(markers_dict.get_item(m.name)._to_dictionary())
            )
        assert len(list(s1.metadata.Markers)) == 8
        for m0_dict, m1_dict in zip(m0_dict_list, m1_dict_list):
            np.testing.assert_equal(m0_dict, m1_dict)

    @zspy_marker
    def test_save_load_horizontal_lines_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        y = 8
        color = "blue"
        linewidth = 2.5
        name = "horizontal_line_test"
        m = hs.plot.markers.HorizontalLines(
            offsets=(y,), color=color, linewidth=linewidth
        )
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    def test_save_load_vertical_lines_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        x = 9
        color = "black"
        linewidth = 3.5
        name = "vertical_line_test"
        m = hs.plot.markers.VerticalLines(
            offsets=(x,), color=color, linewidth=linewidth
        )
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    def test_save_load_lines_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        segments = np.array([[[1, 9], [4, 7]]])
        color = "cyan"
        linewidth = 0.7
        name = "line_segment_test"
        m = hs.plot.markers.Lines(segments=segments, color=color, linewidth=linewidth)
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    def test_save_load_points_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        offsets = [9, 8]
        color = "purple"
        name = "point test"
        m = hs.plot.markers.Points(offsets, color=color)
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    def test_save_load_rectangles_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        offsets = [2, 4]
        widths, heights = 1, 3
        color = "yellow"
        linewidth = 5
        name = "rectangle_test"
        m = hs.plot.markers.Rectangles(
            offsets=offsets,
            widths=widths,
            heights=heights,
            color=color,
            linewidth=linewidth,
        )
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    def test_save_load_text_marker(self, tmp_path, file):
        filename = tmp_path / file
        s = self.s
        offsets = [3, 9.5]
        color = "brown"
        name = "text_test"
        texts = [
            "a text",
        ]
        m = hs.plot.markers.Texts(offsets=offsets, texts=texts, color=color)
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )

    @zspy_marker
    @pytest.mark.parametrize("lazy", [True, False])
    def test_save_load_multidim_navigation_marker(self, tmp_path, file, lazy):
        filename = tmp_path / file
        offsets = np.array([[1, 2, 3], [5, 6, 7]]).T
        name = "test point"
        s = hs.signals.Signal2D(np.arange(300).reshape(3, 10, 10))
        if lazy:
            s = s.as_lazy()
        m = hs.plot.markers.Points(offsets=offsets)
        m.name = name
        s.add_marker(m, permanent=True)
        s.save(filename)
        s1 = hs.load(filename)
        m1 = s1.metadata.Markers.get_item(name)
        np.testing.assert_equal(
            san_dict(m1._to_dictionary()), san_dict(m._to_dictionary())
        )
        for i, (offset_ref, offset) in enumerate(
            zip(offsets.T, m1.get_current_kwargs()["offsets"].T)
        ):
            s1.axes_manager.navigation_axes[0].index = i
            np.testing.assert_allclose(offset_ref, offset)

    def test_load_unknown_marker_type(self):
        # test_marker_bad_marker_type.hdf5 has 5 markers,
        # where one of them has an unknown marker type and raise an error
        fname = TEST_DATA_PATH / "test_marker_bad_marker_type.hdf5"
        with pytest.raises(AttributeError):
            _ = hs.load(fname, reader="HSPY")

    def test_load_missing_y2_value(self):
        # test_marker_point_y2_data_deleted.hdf5 has 5 markers,
        # where one of them is missing the y2 value, however the
        # the point marker only needs the x1 and y1 value to work
        # so this should load
        fname = TEST_DATA_PATH / "test_marker_point_y2_data_deleted.hdf5"
        with pytest.warns(VisibleDeprecationWarning):
            # The binned attribute has been moved from metadata.Signal
            s = hs.load(fname, reader="HSPY")
        assert len(s.metadata.Markers) == 5

    def test_save_variable_length_markers(self, tmp_path):
        fname = tmp_path / "test.hspy"
        rng = np.random.default_rng(0)
        data = np.arange(25 * 25 * 100 * 100).reshape((25, 25, 100, 100))
        s = hs.signals.Signal2D(data)

        offsets = np.empty(s.axes_manager.navigation_shape, dtype=object)
        for ind in np.ndindex(offsets.shape):
            num = rng.integers(3, 10)
            offsets[ind] = rng.random((num, 2)) * 100

        m = hs.plot.markers.Points(
            offsets=offsets,
            color="orange",
        )

        s.plot()
        s.add_marker(m, permanent=True)
        s.save(fname)

        s2 = hs.load(fname)
        s2.plot()

    @zspy_marker
    def test_texts_markers(self, tmp_path, file):
        # h5py doesn't support numpy unicode dtype and when saving ragged
        # array with
        fname = tmp_path / file

        # Create a Signal2D with 1 navigation dimension
        rng = np.random.default_rng(0)
        data = np.ones((5, 100, 100))
        s = hs.signals.Signal2D(data)

        # Create an navigation dependent (ragged) Texts marker
        offsets = np.empty(s.axes_manager.navigation_shape, dtype=object)
        texts = np.empty(s.axes_manager.navigation_shape, dtype=object)

        for index in np.ndindex(offsets.shape):
            i = index[0]
            offsets[index] = rng.random((5, 2))[: i + 2] * 100
            texts[index] = np.array(["a" * (i + 1), "b", "c", "d", "e"][: i + 2])

        m = hs.plot.markers.Texts(
            offsets=offsets,
            texts=texts,
            sizes=3,
            facecolor="black",
        )

        s.add_marker(m, permanent=True)
        s.plot()
        s.save(fname)

        s2 = hs.load(fname)

        m_texts = m.kwargs["texts"]
        m2_texts = s2.metadata.Markers.Texts.kwargs["texts"]

        for index in np.ndindex(m_texts.shape):
            np.testing.assert_equal(m_texts[index], m2_texts[index])


@zspy_marker
@pytest.mark.parametrize("use_list", [True, False])
def test_saving_ragged_array_string(tmp_path, file, use_list):
    # h5py doesn't support numpy unicode dtype and when saving ragged
    # array, we need to change the array dtype
    fname = tmp_path / file

    string_data = np.empty((5,), dtype=object)
    for index in np.ndindex(string_data.shape):
        i = index[0]
        data = np.array(["a" * (i + 1), "b", "c", "d", "e"][: i + 2])
        if use_list:
            data = data.tolist()
        string_data[index] = data

    s = hs.signals.BaseSignal(string_data, ragged=True)
    s.save(fname)

    s2 = hs.load(fname)
    for index in np.ndindex(s.data.shape):
        np.testing.assert_equal(s.data[index], s2.data[index])


@zspy_marker
def test_saving_ragged_array_single_string(tmp_path, file):
    fname = tmp_path / file

    string_data = np.empty((2, 5), dtype=object)
    for i, index in enumerate(np.ndindex(string_data.shape)):
        string_data[index] = "a" * (i + 1)

    s = hs.signals.BaseSignal(string_data, ragged=True)

    s.save(fname, overwrite=True)

    s2 = hs.load(fname)
    for index in np.ndindex(s.data.shape):
        np.testing.assert_equal(s.data[index], s2.data[index])


@zspy_marker
@pytest.mark.parametrize("lazy", [True, False])
def test_save_load_model(tmp_path, file, lazy):
    filename = tmp_path / file
    s = hs.signals.Signal1D(np.ones((10, 10, 10, 10)))
    if lazy:
        s = s.as_lazy()
    m = s.create_model()
    m.append(hs.model.components1D.Gaussian())
    m.store("test")
    s.save(filename)
    signal2 = hs.load(filename)
    m2 = signal2.models.restore("test")
    assert m.signal == m2.signal


@pytest.mark.parametrize("compression", (None, "gzip", "lzf"))
def test_compression(compression, tmp_path):
    s = hs.signals.Signal1D(np.ones((3, 3)))
    s.save(tmp_path / "test_compression.hspy", compression=compression)
    _ = hs.load(tmp_path / "test_compression.hspy")


def test_strings_from_py2():
    exspy = pytest.importorskip("exspy", reason="exspy not installed")
    s = exspy.data.EDS_TEM_FePt_nanoparticles()
    assert isinstance(s.metadata.Sample.elements, list)


@zspy_marker
def test_save_ragged_array(tmp_path, file):
    a = np.array([0, 1])
    b = np.array([0, 1, 2])
    s = hs.signals.BaseSignal(np.array([a, b], dtype=object), ragged=True)
    assert s.ragged
    fname = tmp_path / file
    s.save(fname)
    s1 = hs.load(fname)
    assert s1.ragged
    for i in range(len(s.data)):
        np.testing.assert_allclose(s.data[i], s1.data[i])
    assert s.__class__ == s1.__class__


@zspy_marker
@pytest.mark.parametrize("nav_dim", [1, 2])
@pytest.mark.parametrize("lazy", [True, False])
def test_save_ragged_dim(tmp_path, file, nav_dim, lazy):
    file = f"nav{nav_dim}_" + file
    rng = np.random.default_rng(0)
    nav_shape = np.arange(10, 10 * (nav_dim + 1), step=10)
    data = np.empty(nav_shape, dtype=object)
    for ind in np.ndindex(data.shape):
        num = rng.integers(3, 10)
        data[ind] = rng.random((num, 2)) * 100

    s = hs.signals.BaseSignal(data, ragged=True)
    if lazy:
        s = s.as_lazy()
    assert s.axes_manager.navigation_dimension == nav_dim
    np.testing.assert_allclose(s.axes_manager.navigation_shape, nav_shape[::-1])
    assert s.data.ndim == nav_dim
    np.testing.assert_allclose(s.data.shape, nav_shape)

    filename = tmp_path / file
    if ".hspy" in file and nav_dim == 1 and lazy:
        with pytest.raises(ValueError):
            s.save(filename)

    else:
        s.save(filename)
        s2 = hs.load(filename, lazy=lazy)
        assert s.axes_manager.navigation_shape == s2.axes_manager.navigation_shape
        assert s.data.shape == s2.data.shape
        if lazy:
            assert isinstance(s2.data, da.Array)

        for indices in np.ndindex(s.data.shape):
            np.testing.assert_allclose(s.data[indices], s2.data[indices])


def test_load_missing_extension(caplog):
    path = TEST_DATA_PATH / "hspy_ext_missing.hspy"
    with pytest.warns(UserWarning):
        s = hs.load(path)
    assert "This file contains a signal provided by the hspy_ext_missing" in caplog.text
    with pytest.raises(ImportError):
        _ = s.models.restore("a")


@zspy_marker
def test_save_chunks_signal_metadata(tmp_path, file):
    N = 10
    dim = 3
    s = hs.signals.Signal1D(np.arange(N**dim).reshape([N] * dim))
    s.navigator = s.sum(-1)
    s.change_dtype("float")
    s.decomposition()

    filename = tmp_path / file
    chunks = (5, 5, 10)
    s.save(filename, chunks=chunks)
    s2 = hs.load(filename, lazy=True)
    assert tuple([c[0] for c in s2.data.chunks]) == chunks


@zspy_marker
def test_chunking_saving_lazy(tmp_path, file):
    s = hs.signals.Signal2D(da.zeros((50, 100, 100))).as_lazy()
    s.data = s.data.rechunk([50, 25, 25])

    filename = tmp_path / "test_chunking_saving_lazy.hspy"
    filename2 = tmp_path / "test_chunking_saving_lazy_chunks_True.hspy"
    filename3 = tmp_path / "test_chunking_saving_lazy_chunks_specified.hspy"
    s.save(filename)
    s1 = hs.load(filename, lazy=True)
    assert s.data.chunks == s1.data.chunks

    # with chunks=True, use h5py chunking
    s.save(filename2, chunks=True)
    s2 = hs.load(filename2, lazy=True)
    assert tuple([c[0] for c in s2.data.chunks]) == (7, 25, 25)

    # specify chunks
    chunks = (50, 10, 10)
    s.save(filename3, chunks=chunks)
    s3 = hs.load(filename3, lazy=True)
    assert tuple([c[0] for c in s3.data.chunks]) == chunks


@pytest.mark.parametrize("show_progressbar", (True, False))
@zspy_marker
def test_saving_show_progressbar(tmp_path, file, show_progressbar):
    s = hs.signals.Signal2D(da.zeros((50, 100, 100))).as_lazy()

    filename = tmp_path / file

    s.save(filename, show_progressbar=show_progressbar)


def test_saving_close_file(tmp_path):
    # Setup that we will reopen
    s = hs.signals.Signal1D(da.zeros((10, 100))).as_lazy()
    fname = tmp_path / "test.hspy"
    s.save(fname, close_file=True)

    s2 = hs.load(fname, lazy=True, mode="a")
    assert get_file_handle(s2.data) is not None
    s2.save(fname, close_file=True, overwrite=True)
    assert get_file_handle(s2.data) is None

    s3 = hs.load(fname, lazy=True, mode="a")
    assert get_file_handle(s3.data) is not None
    s3.save(fname, close_file=False, overwrite=True)
    assert get_file_handle(s3.data) is not None
    s3.close_file()
    assert get_file_handle(s3.data) is None

    s4 = hs.load(fname, lazy=True)
    assert get_file_handle(s4.data) is not None
    with pytest.raises(OSError):
        s4.save(fname, overwrite=True)


@zspy_marker
def test_saving_overwrite_data(tmp_path, file):
    s = hs.signals.Signal1D(da.zeros((10, 100))).as_lazy()
    kwds = dict(close_file=True) if file == "test.hspy" else {}
    fname = tmp_path / file
    s.save(fname, **kwds)

    s2 = hs.load(fname, lazy=True, mode="a")
    s2.axes_manager[0].units = "nm"
    s2.data = da.ones((10, 100))
    s2.save(fname, overwrite=True, write_dataset=False)

    s3 = hs.load(fname)
    assert s3.axes_manager[0].units == "nm"
    # Still old data
    np.testing.assert_allclose(s3.data, np.zeros((10, 100)))

    s4 = hs.load(fname)
    s4.data = da.ones((10, 100))
    if file == "test.hspy":
        # try with opening non-lazily to check opening mode
        # only relevant for hdf5 file
        with pytest.raises(ValueError):
            s4.save(fname, overwrite=True, write_dataset=False, mode="w")

    s4.save(fname, overwrite=True, write_dataset=True)
    # make sure we can open it after, file haven't been corrupted
    _ = hs.load(fname)

    # now new data
    @zspy_marker
    def test_chunking_saving_lazy_specify(self, tmp_path, file):
        filename = tmp_path / file
        s = hs.signals.Signal2D(da.zeros((50, 100, 100))).as_lazy()
        # specify chunks
        chunks = (50, 10, 10)
        s.data = s.data.rechunk([50, 25, 25])
        s.save(filename, chunks=chunks)
        s1 = hs.load(filename, lazy=True)
        assert tuple([c[0] for c in s1.data.chunks]) == chunks


@pytest.mark.parametrize("target_size", (1e6, 1e7))
def test_get_signal_chunks(target_size):
    shape = (2, 150, 3, 200, 1, 600, 1)
    chunks = get_signal_chunks(
        shape=shape, dtype=np.int64, signal_axes=(2, 3), target_size=target_size
    )
    assert np.prod(chunks) * 8 < target_size
    # The chunks must be smaller or equal that the corresponding sizes
    assert (np.array(chunks) <= np.array(shape)).all()


def test_get_signal_chunks_big_signal():
    # One signal exceeds the target size
    shape = (10, 1000, 5, 1000)
    chunks = get_signal_chunks(
        shape=shape, dtype=np.int32, signal_axes=(1, 3), target_size=1e6
    )
    # The chunks must be smaller or equal that the corresponding sizes
    assert chunks == (1, 1000, 1, 1000)


def test_get_signal_chunks_small_dataset():
    # Whole dataset fits in one chunk
    shape = (10, 10, 2, 2)
    chunks = get_signal_chunks(
        shape=shape, dtype=np.int32, signal_axes=(2, 3), target_size=1e6
    )
    # The chunks must be smaller or equal that the corresponding sizes
    assert chunks == shape


@zspy_marker
def test_error_saving(tmp_path, file):
    filename = tmp_path / file
    s = hs.signals.Signal1D(np.arange(10))

    with pytest.raises(ValueError):
        s.save(filename, write_dataset="unsupported_type")
        assert not s.metadata.Signal.has_item("record_by")


def test_more_recent_version_warning(tmp_path):
    filename = tmp_path / "test.hspy"
    s = hs.signals.Signal1D(np.arange(10))
    s.save(filename)

    with h5py.File(filename, mode="a") as f:
        f.attrs["file_format_version"] = "99999999"

    with pytest.warns(UserWarning):
        s2 = hs.load(filename)
    np.testing.assert_allclose(s.data, s2.data)


@zspy_marker
def test_save_metadata_empty_array(tmp_path, file):
    filename = tmp_path / "test.hspy"

    s = hs.signals.Signal1D(np.arange(100 * 1024).reshape(100, 1024))
    s.metadata.set_item("Sample.xray_lines", np.array([]))

    s.save(filename)
    s2 = hs.load(filename)

    np.testing.assert_allclose(
        s.metadata.Sample.xray_lines, s2.metadata.Sample.xray_lines
    )
    np.testing.assert_allclose(s.data, s2.data)


@zspy_marker
def test_save_empty_signal(tmp_path, file):
    filename = tmp_path / "test.hspy"

    s = hs.signals.Signal1D([])

    s.save(filename)
    s2 = hs.load(filename)

    np.testing.assert_allclose(s.data, s2.data)



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

import importlib

import pytest

PLUGIN_LAZY_NOT_IMPLEMENTED = [
    # "bruker", # SPX only
    "dens",
    "digitalsurf",
    "impulse",
    "jobinyvon",
    "msa",
    "netcdf",
    "phenom",
    "protochips",
    "renishaw",
    "trivista",
]


@pytest.mark.parametrize("plugin", PLUGIN_LAZY_NOT_IMPLEMENTED)
def test_lazy_not_implemented(plugin):
    fname = "fname"
    if plugin == "bruker":
        fname = "fname.spx"
    if plugin == "phenom":
        pytest.importorskip("tifffile")
    if plugin == "netcdf":
        pytest.importorskip("scipy")

    file_reader = importlib.import_module(f"rsciio.{plugin}").file_reader

    with pytest.raises(NotImplementedError):
        file_reader(fname, lazy=True)



import copy
import importlib
from pathlib import Path

import numpy as np
import pytest

from rsciio import __version__, msa
from rsciio.utils._tests import assert_deep_almost_equal

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")


TEST_DATA_PATH = Path(__file__).parent / "data" / "msa"


example1_TEM = {
    "Detector": {
        "EELS": {
            "collection_angle": 3.4,
            "collection_angle_units": "mR",
            "dwell_time": 100.0,
            "dwell_time_units": "ms",
        }
    },
    "beam_current": 12.345,
    "beam_current_units": "nA",
    "beam_energy": 120.0,
    "beam_energy_units": "kV",
    "convergence_angle": 1.5,
    "convergence_angle_units": "mR",
}

example1_metadata = {
    "Acquisition_instrument": {"TEM": example1_TEM},
    "General": {
        "original_filename": "example1.msa",
        "title": "NIO EELS OK SHELL",
        "date": "1991-10-01",
        "time": "12:00:00",
        "FileIO": {
            "0": {
                "operation": "load",
                "hyperspy_version": hs.__version__,
                "io_plugin": "rsciio.msa",
            }
        },
    },
    "Sample": {"thickness": 50.0, "thickness_units": "nm"},
    "Signal": {"quantity": "Counts (Intensity)", "signal_type": "EELS"},
    "_HyperSpy": {
        "Folding": {
            "original_axes_manager": None,
            "unfolded": False,
            "original_shape": None,
            "signal_unfolded": False,
        }
    },
}
minimum_md_om = {
    "COMMENT": "File created by HyperSpy version 1.1.2+dev",
    "DATATYPE": "Y",
    "DATE": "",
    "FORMAT": "EMSA/MAS Spectral Data File",
    "NCOLUMNS": 1.0,
    "NPOINTS": 1.0,
    "OFFSET": 0.0,
    "OWNER": "",
    "SIGNALTYPE": "",
    "TIME": "",
    "TITLE": "",
    "VERSION": "1.0",
    "XLABEL": "",
    "XPERCHAN": 1.0,
    "XUNITS": "",
}


example1_parameters = {
    "BEAMDIAM -nm": 100.0,
    "BEAMKV   -kV": 120.0,
    "CHOFFSET": -168.0,
    "COLLANGLE-mR": 3.4,
    "CONVANGLE-mR": 1.5,
    "DATATYPE": "XY",
    "DATE": "01-OCT-1991",
    "DWELLTIME-ms": 100.0,
    "ELSDET": "SERIAL",
    "EMISSION -uA": 5.5,
    "FORMAT": "EMSA/MAS Spectral Data File",
    "MAGCAM": 100.0,
    "NCOLUMNS": 1.0,
    "NPOINTS": 20.0,
    "OFFSET": 520.13,
    "OPERMODE": "IMAG",
    "OWNER": "EMSA/MAS TASK FORCE",
    "PROBECUR -nA": 12.345,
    "SIGNALTYPE": "ELS",
    "THICKNESS-nm": 50.0,
    "TIME": "12:00",
    "TITLE": "NIO EELS OK SHELL",
    "VERSION": "1.0",
    "XLABEL": "Energy",
    "XPERCHAN": 3.1,
    "XUNITS": "eV",
    "YLABEL": "Counts",
    "YUNITS": "Intensity",
}

example2_TEM = {
    "Detector": {
        "EDS": {
            "EDS_det": "SIWLS",
            "azimuth_angle": 90.0,
            "azimuth_angle_units": "dg",
            "elevation_angle": 20.0,
            "elevation_angle_units": "dg",
            "energy_resolution_MnKa": 130.0,
            "live_time": 100.0,
            "live_time_units": "s",
            "real_time": 150.0,
            "real_time_units": "s",
        }
    },
    "beam_current": 12.345,
    "beam_current_units": "nA",
    "beam_energy": 120.0,
    "beam_energy_units": "kV",
    "Stage": {"tilt_alpha": 45.0, "tilt_alpha_units": "dg"},
}

example2_metadata = {
    "Acquisition_instrument": {"TEM": example2_TEM},
    "General": {
        "original_filename": "example2.msa",
        "title": "NIO Windowless Spectra OK NiL",
        "date": "1991-10-01",
        "time": "12:00:00",
        "FileIO": {
            "0": {
                "operation": "load",
                "hyperspy_version": hs.__version__,
                "io_plugin": "rsciio.msa",
            }
        },
    },
    "Sample": {"thickness": 50.0, "thickness_units": "nm"},
    "Signal": {"quantity": "X-RAY INTENSITY (Intensity)", "signal_type": "EDS_TEM"},
    "_HyperSpy": {
        "Folding": {
            "original_axes_manager": None,
            "unfolded": False,
            "original_shape": None,
            "signal_unfolded": False,
        }
    },
}

example2_parameters = {
    "ALPHA-1": "3.1415926535",
    "AZIMANGLE-dg": 90.0,
    "BEAMDIAM -nm": 100.0,
    "BEAMKV   -kV": 120.0,
    "CHOFFSET": -20.0,
    "COMMENT": "The next two lines are User Defined Keywords and values",
    "DATATYPE": "Y",
    "DATE": "01-OCT-1991",
    "EDSDET": "SIWLS",
    "ELEVANGLE-dg": 20.0,
    "EMISSION -uA": 5.5,
    "FORMAT": "EMSA/MAS Spectral Data File",
    "LIVETIME  -s": 100.0,
    "MAGCAM": 100.0,
    "NCOLUMNS": 1.0,
    "NPOINTS": 80.0,
    "OFFSET": 200.0,
    "OPERMODE": "IMAG",
    "OWNER": "EMSA/MAS TASK FORCE",
    "PROBECUR -nA": 12.345,
    "REALTIME  -s": 150.0,
    "RESTMASS": "511.030",
    "SIGNALTYPE": "EDS",
    "SOLIDANGL-sR": "0.13",
    "TACTLYR  -cm": 0.3,
    "TAUWIND  -cm": 2e-06,
    "TBEWIND  -cm": 0.0,
    "TDEADLYR -cm": 1e-06,
    "THICKNESS-nm": 50.0,
    "TIME": "12:00",
    "TITLE": "NIO Windowless Spectra OK NiL",
    "VERSION": "1.0",
    "XLABEL": "X-RAY ENERGY",
    "XPERCHAN": 10.0,
    "XPOSITION": 123.0,
    "XTILTSTGE-dg": 45.0,
    "XUNITS": "eV",
    "YLABEL": "X-RAY INTENSITY",
    "YPOSITION": 456.0,
    "YTILTSTGE-dg": 20.0,
    "YUNITS": "Intensity",
    "ZPOSITION": 0.0,
}


class TestExample1:
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example1.msa")
        # delete timestamp from metadata since it's runtime dependent
        del self.s.metadata.General.FileIO.Number_0.timestamp

    def test_data(self):
        assert [
            4066.0,
            3996.0,
            3932.0,
            3923.0,
            5602.0,
            5288.0,
            7234.0,
            7809.0,
            4710.0,
            5015.0,
            4366.0,
            4524.0,
            4832.0,
            5474.0,
            5718.0,
            5034.0,
            4651.0,
            4613.0,
            4637.0,
            4429.0,
            4217.0,
        ] == self.s.data.tolist()

    def test_parameters(self):
        assert example1_parameters == self.s.original_metadata.as_dictionary()

    def test_metadata(self):
        assert_deep_almost_equal(self.s.metadata.as_dictionary(), example1_metadata)

    def test_write_load_cycle(self, tmp_path):
        fname2 = tmp_path / "example1-export.msa"
        self.s.save(fname2)
        s2 = hs.load(fname2)
        # delete timestamp from metadata since it's runtime dependent
        del s2.metadata.General.FileIO.Number_0.timestamp
        del self.s.metadata.General.FileIO.Number_1
        if "timestamp" in self.s.metadata.General.FileIO.Number_0:
            del self.s.metadata.General.FileIO.Number_0.timestamp
        assert s2.metadata.General.original_filename == "example1-export.msa"
        s2.metadata.General.original_filename = "example1.msa"
        assert_deep_almost_equal(
            self.s.metadata.as_dictionary(), s2.metadata.as_dictionary()
        )


class TestExample1WrongDate:
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example1_wrong_date_empty_field.msa")
        # delete timestamp from metadata since it's runtime dependent
        del self.s.metadata.General.FileIO.Number_0.timestamp

    def test_metadata(self):
        md = copy.copy(example1_metadata)
        del md["General"]["date"]
        del md["General"]["time"]
        md["General"]["original_filename"] = "example1_wrong_date_empty_field.msa"
        assert_deep_almost_equal(self.s.metadata.as_dictionary(), md)


class TestExample2:
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "example2.msa")
        # delete timestamp from metadata since it's runtime dependent
        del self.s.metadata.General.FileIO.Number_0.timestamp

    def test_data(self):
        assert [
            65.82,
            67.872,
            65.626,
            68.762,
            71.395,
            74.996,
            78.132,
            78.055,
            77.861,
            84.598,
            83.088,
            85.372,
            89.786,
            93.464,
            93.387,
            97.452,
            109.96,
            111.08,
            119.64,
            128.77,
            138.38,
            152.35,
            176.01,
            192.12,
            222.12,
            254.22,
            281.55,
            328.33,
            348.92,
            375.33,
            385.51,
            389.54,
            378.77,
            353.8,
            328.91,
            290.07,
            246.09,
            202.73,
            176.47,
            137.64,
            119.56,
            106.4,
            92.496,
            96.213,
            94.664,
            101.13,
            114.57,
            118.82,
            131.68,
            145.04,
            165.44,
            187.51,
            207.49,
            238.04,
            269.71,
            301.46,
            348.65,
            409.36,
            475.3,
            554.51,
            631.64,
            715.19,
            793.44,
            847.99,
            872.97,
            862.59,
            834.87,
            778.49,
            688.63,
            599.39,
            495.39,
            403.48,
            312.88,
            237.34,
            184.14,
            129.86,
            101.59,
            80.107,
            58.657,
            49.442,
        ] == self.s.data.tolist()

    def test_parameters(self):
        assert example2_parameters == self.s.original_metadata.as_dictionary()

    def test_metadata(self):
        expected_TEM = copy.deepcopy(example2_TEM)
        if importlib.util.find_spec("exspy") is None:
            del expected_TEM["Detector"]["EDS"]["energy_resolution_MnKa"]
        assert (
            self.s.metadata.Acquisition_instrument.TEM.as_dictionary() == expected_TEM
        )

    def test_write_load_cycle(self, tmp_path):
        fname2 = tmp_path / "example2-export.msa"
        self.s.save(fname2)
        s2 = hs.load(fname2)
        assert s2.metadata.General.original_filename == "example2-export.msa"
        s2.metadata.General.original_filename = "example2.msa"
        # delete timestamp from metadata since it's runtime dependent
        del s2.metadata.General.FileIO.Number_0.timestamp
        del self.s.metadata.General.FileIO.Number_1
        assert_deep_almost_equal(
            self.s.metadata.as_dictionary(), s2.metadata.as_dictionary()
        )
        np.testing.assert_allclose(self.s.data, s2.data)


def test_DATATYPE_Y_NCOLUMNS5():
    s = hs.load(TEST_DATA_PATH / "example2.msa")
    assert s.original_metadata["DATATYPE"] == "Y"
    assert s.original_metadata["NCOLUMNS"] == 1.0

    s2 = hs.load(TEST_DATA_PATH / "example2_NCOLUMNS5.msa")
    assert s2.original_metadata["DATATYPE"] == "Y"
    assert s2.original_metadata["NCOLUMNS"] == 5.0
    np.testing.assert_allclose(s.data, s2.data)


def test_minimum_metadata_example():
    s = hs.load(TEST_DATA_PATH / "minimum_metadata.msa")
    assert minimum_md_om == s.original_metadata.as_dictionary()


class TestSignalType:
    def setup_method(self, method):
        self.s = hs.load(TEST_DATA_PATH / "minimum_metadata.msa")
        # delete timestamp from metadata since it's runtime dependent
        del self.s.metadata.General.FileIO.Number_0.timestamp

    @pytest.mark.parametrize(
        "signaltype", ("EDS_SEM", "EDS_TEM", "CL", "EELS", "PES", "brian")
    )
    def test_write_minimum_metadata_signaltype(self, signaltype, tmp_path):
        fname2 = tmp_path / "example-min-export.msa"
        self.s.metadata.Signal.signal_type = signaltype
        self.s.save(fname2)
        s2 = hs.load(fname2)
        if "EDS" in signaltype:
            assert s2.original_metadata["SIGNALTYPE"] == "EDS"
        elif signaltype == "CL":
            assert s2.original_metadata["SIGNALTYPE"] == "CLS"
        elif signaltype == "EELS":
            assert s2.original_metadata["SIGNALTYPE"] == "ELS"
        elif signaltype not in [
            "EDS",
            "WDS",
            "ELS",
            "AES",
            "PES",
            "XRF",
            "CLS",
            "GAM",
        ]:
            assert s2.original_metadata["SIGNALTYPE"] == ""
        else:
            assert s2.original_metadata["SIGNALTYPE"] in [
                "EDS",
                "WDS",
                "ELS",
                "AES",
                "PES",
                "XRF",
                "CLS",
                "GAM",
            ]


def test_iso_compliance(tmp_path):
    # Example file from the 2022 revision: https://www.iso.org/standard/78268.html
    s = hs.load(TEST_DATA_PATH / "ISO_22029_2022_compliance.msa")
    assert s.original_metadata.as_dictionary() == {
        "FORMAT": "EMSA/MAS Spectral Data File",
        "VERSION": "TC202v2.0",
        "TITLE": "NIO EELS OK SHELL",
        "DATE": "01-OCT-1991",
        "TIME": "12:00",
        "OWNER": "EMSA/MAS TASK FORCE",
        "NPOINTS": 21.0,
        "NCOLUMNS": 1.0,
        "XUNITS": "Energy Loss (eV)",
        "YUNITS": "Intensity",
        "DATATYPE": "XY",
        "XPERCHAN": 3.1,
        "OFFSET": 520.13,
        "CHOFFSET": -168.0,
        "SIGNALTYPE": "ELS",
        "XLABEL": "Energy",
        "YLABEL": "Counts",
        "BEAMKV": 120.0,
        "EMISSION": 5.5,
        "PROBECUR": 12.345,
        "BEAMDIAM": 100.0,
        "MAGCAM": 100.0,
        "CONVANGLE": 1.5,
        "COLLANGLE": 3.4,
        "OPERMODE": "IMAG",
        "THICKNESS": 50.0,
        "DWELLTIME": 100.0,
        "ELSDET": "SERIAL",
    }
    assert s.metadata.Signal.signal_type == "EELS"
    assert s.metadata.General.title == "NIO EELS OK SHELL"

    s2 = hs.load(TEST_DATA_PATH / "ISO_22029_2022_compliance_scientific_notation.msa")
    np.testing.assert_allclose(s.data, s2.data)

    s3 = hs.load(TEST_DATA_PATH / "ISO_22029_2022_compliance_title_multiple_line.msa")
    assert (
        s3.metadata.General.title
        == "NIO EELS OK SHELL - Second line of the title - Third line of the title"
    )
    # Title is longer than 64 characters, multiple TITLE keywords will be used to store the full title.

    fname = tmp_path / "ISO_22029_2022_compliance_title_multiple_line_export.msa"
    s3.save(fname)
    s4 = hs.load(fname)
    assert s4.metadata.General.title == s3.metadata.General.title

    s5 = hs.load(TEST_DATA_PATH / "ISO_22029_2022_compliance_XY_NCOLUMNS2.msa")
    np.testing.assert_allclose(s5.data, s.data)


def test_time_with_seconds():
    # Example file with time including seconds, which is not compliant with ISO standard,
    # but it is found in some files.
    s = hs.load(TEST_DATA_PATH / "example1_with_seconds.msa")
    assert s.metadata.General.time == "12:00:00"


def test_parsed_version(tmp_path):
    s = msa.file_reader(TEST_DATA_PATH / "minimum_metadata.msa")[0]
    fname = tmp_path / "example_with_version.msa"
    _ = s["original_metadata"].pop("COMMENT")
    msa.file_writer(fname, s)

    # drop the revision hash from the setuptools_scm version if its there
    # as it's long enough to split across lines.
    version_parts = __version__.split(".")
    nv = 3 if len(version_parts) == 3 else 4
    v = ".".join(version_parts[:nv])
    with open(fname, "r") as fi:
        assert v in fi.read()



# -*- coding: utf-8 -*-
# Copyright 2007-2026 The HyperSpy developers
#
# This file is part of RosettaSciIO.
#
# RosettaSciIO is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# RosettaSciIO is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RosettaSciIO. If not, see <https://www.gnu.org/licenses/#GPL>.

from pathlib import Path

import numpy as np
import pytest

hs = pytest.importorskip("hyperspy.api", reason="hyperspy not installed")
t = pytest.importorskip("traits.api", reason="traits not installed")
pytest.importorskip("h5py", reason="h5py not installed")

TEST_DATA_PATH = Path(__file__).parent / "data" / "emd"


def test_all_but_4D():
    filename = TEST_DATA_PATH / "Si100_2D_3D_DPC_potential_2slices.emd"
    s = hs.load(filename)

    assert len(s) == 4
    # DPC: navigation dimension are DPC and depth
    assert s[0].data.shape == (2, 2, 22, 22)
    # annular detectors: navigation dimension is depth
    assert s[1].data.shape == (2, 22, 22)
    # virtual detectors: navigation dimension are depth and scattering angle
    assert s[2].data.shape == (4, 16, 16)
    # projected potential: navigation dimension is number of slices
    assert s[3].data.shape == (2, 18, 22, 22)

    def check_depth_axis(axis):
        assert "depth" in axis.name
        assert axis.units == "Å"
        np.testing.assert_allclose(axis.scale, 2.715)
        np.testing.assert_allclose(axis.offset, 2.715)
        np.testing.assert_allclose(axis.size, 2)

    def check_signal_axes(signal_axes):
        assert signal_axes[0].name == "R_x"
        assert signal_axes[1].name == "R_y"
        for axis in signal_axes:
            assert axis.units == "nm"
            np.testing.assert_allclose(axis.scale, 0.25)
            np.testing.assert_allclose(axis.offset, 0)
            np.testing.assert_allclose(axis.size, 22)

    for i in [0, 1, 3]:
        # signal axis
        check_signal_axes(s[i].axes_manager.signal_axes)
        # depth axis
        check_depth_axis(s[i].axes_manager.navigation_axes[-1])

    # DPC axis
    axis = s[0].axes_manager[0]
    assert axis.name == t.Undefined
    assert axis.units == t.Undefined
    np.testing.assert_allclose(axis.scale, 1)
    np.testing.assert_allclose(axis.offset, 0)
    np.testing.assert_allclose(axis.size, 2)

    axis = s[2].axes_manager[0]
    assert axis.name == "R_z"
    assert axis.units == "nm"
    np.testing.assert_allclose(axis.scale, 1.3575)
    np.testing.assert_allclose(axis.offset, 0)
    np.testing.assert_allclose(axis.size, 4)

    # signal axes
    signal_axes = s[2].axes_manager.signal_axes
    assert signal_axes[0].name == "R_x"
    assert signal_axes[1].name == "R_y"
    for axis in signal_axes:
        assert axis.units == "nm"
        np.testing.assert_allclose(axis.scale, 0.339375)
        np.testing.assert_allclose(axis.offset, 0)
        np.testing.assert_allclose(axis.size, 16)


def test_depth_axis_zStart():
    filename = TEST_DATA_PATH / "Si100_1x1x3-zStart5.43.emd"
    s = hs.load(filename)
    axis = s.axes_manager[1]

    assert "depth" in axis.name
    assert axis.units == "Å"
    np.testing.assert_allclose(axis.scale, 5.43)
    np.testing.assert_allclose(axis.offset, 5.43)
    np.testing.assert_allclose(axis.size, 3)

    # non-uniform depth axis, not supported yet
    # only the depth of the last image differ from others
    filename = TEST_DATA_PATH / "Si100_1x1x3-zStart6.7875.emd"
    s = hs.load(filename)
    axis = s.axes_manager[1]

    assert "depth" in axis.name
    assert axis.units == t.Undefined
    np.testing.assert_allclose(axis.scale, 1)
    np.testing.assert_allclose(axis.offset, 0)
    np.testing.assert_allclose(axis.size, 2)


def test_all_but_4D_no_stack():
    filename = TEST_DATA_PATH / "Si100_2D_3D_DPC_potential_2slices.emd"
    s = hs.load(filename, stack_group=False)
    assert len(s) == 7

    s2 = hs.load(filename, stack_group=True)
    np.testing.assert_allclose(np.stack([s[0].data, s[1].data]), s2[0].data)


@pytest.mark.parametrize(
    "dataset_path",
    [
        "4DSTEM_simulation/data/realslices/annular_detector_depth0000/realslice",
        "4DSTEM_simulation/data/realslices/virtual_detector_depth0000/realslice",
        "4DSTEM_simulation/data/realslices/DPC_CoM_depth0000/realslice",
        "4DSTEM_simulation/data/realslices/ppotential/realslice",
    ],
)
def test_load_single_dataset(dataset_path):
    filename = TEST_DATA_PATH / "Si100_2D_3D_DPC_potential_2slices.emd"
    s = hs.load(filename, dataset_path=dataset_path)

    assert isinstance(s, hs.signals.Signal2D)


def test_load_specific_datasets():
    filename = TEST_DATA_PATH / "Si100_2D_3D_DPC_potential_2slices.emd"
    dataset_path = [
        "4DSTEM_simulation/data/realslices/annular_detector_depth0000/realslice",
        "4DSTEM_simulation/data/realslices/virtual_detector_depth0000/realslice",
    ]
    s = hs.load(filename, dataset_path=dataset_path)

    assert len(s) == 2


@pytest.mark.parametrize("lazy", (True, False))
def test_3D_only(lazy):
    filename = TEST_DATA_PATH / "Si100_3D.emd"
    s = hs.load(filename, lazy=lazy)
    if lazy:
        s.compute(close_file=True)

    assert s.data.shape == (37, 22, 22)

    # scattering angle axis
    axis = s.axes_manager[0]
    assert axis.name == "bin_outer_angle"
    assert axis.units == "mrad"
    np.testing.assert_allclose(axis.scale, 0.001, rtol=1e-6)
    np.testing.assert_allclose(axis.offset, 0.0005)
    np.testing.assert_allclose(axis.size, 37)

    # signal axes
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "R_x"
    assert signal_axes[1].name == "R_y"
    for axis in signal_axes:
        assert axis.units == "nm"
        np.testing.assert_allclose(axis.scale, 0.25)
        np.testing.assert_allclose(axis.offset, 0)
        np.testing.assert_allclose(axis.size, 22)


def test_non_square_3D():
    filename = TEST_DATA_PATH / "Si100_2x1x1_3D.emd"
    s = hs.load(filename)

    assert s.data.shape == (22, 44)

    # signal axes
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "R_x"
    assert signal_axes[1].name == "R_y"
    assert signal_axes[0].size == 44
    assert signal_axes[1].size == 22
    for axis in signal_axes:
        assert axis.units == "nm"
        np.testing.assert_allclose(axis.scale, 0.25)
        np.testing.assert_allclose(axis.offset, 0)


@pytest.mark.parametrize("lazy", (True, False))
def test_4D(lazy):
    filename = TEST_DATA_PATH / "Si100_4D.emd"
    s = hs.load(filename, lazy=lazy)
    if lazy:
        s.compute(close_file=True)
    assert s.data.shape == (2, 11, 11, 8, 8)

    # navigation x, y axes
    navigation_axes = s.axes_manager.navigation_axes
    assert navigation_axes[0].name == "R_y"
    assert navigation_axes[1].name == "R_x"
    for axis in navigation_axes[:2]:
        assert axis.units == "nm"
        np.testing.assert_allclose(axis.scale, 0.5)
        np.testing.assert_allclose(axis.offset, 0)
        np.testing.assert_allclose(axis.size, 11)

    axis = s.axes_manager[2]
    assert axis.name == "CBED_array_depth"
    assert axis.units == "Å"
    # Needs metadata to know these values?
    np.testing.assert_allclose(axis.scale, 2.715)
    np.testing.assert_allclose(axis.offset, 2.715)
    np.testing.assert_allclose(axis.size, 2)

    # signal axes
    signal_axes = s.axes_manager.signal_axes
    assert signal_axes[0].name == "Q_y"
    assert signal_axes[1].name == "Q_x"
    assert signal_axes[0].size == 8
    assert signal_axes[1].size == 8
    for axis in signal_axes:
        assert axis.units == "1 / nm"
        np.testing.assert_allclose(axis.scale, 0.18416205)
        np.testing.assert_allclose(axis.offset, -0.73664826)
